import datetime
import json
import unittest
from copy import deepcopy

from metaform import Dict, align, convert, converters, formatize, metaplate, normalize, template


class TestMain(unittest.TestCase):

    def setUp(self):
        self.key = 'something'
        self.value = '1,234'

    def test_simple_conversion(self):
        schema = {
            '*': "IN:mindey/thing|lambda _: _.replace(',','')"
        }

        expect = ('IN:mindey/thing', '1234')

        self.assertEqual(convert(self.key, self.value, schema), expect)

    def test_conversion(self):
        schema = {
            '*': "https://www.wikidata.org/wiki/Q82799|lambda _: _.replace(',','')"
        }

        expect = ('https://www.wikidata.org/wiki/Q82799', '1234')

        self.assertEqual(convert(self.key, self.value, schema), expect)

    def test_normalization(self):
        data = [
            {
                'address': {'number': 14, 'street': 'Leonardo str.'},
                'children': [{'age': 1, 'name': 'Mike'}, {'age': 15, 'name': 'Tom'}],
                'name': 'Max'
            },
            {
                'address': {'number': 1, 'street': 'Nexus str.'},
                'children': [{'age': 1, 'name': 'Deli'}, {'age': 7, 'name': 'Miki'}],
                'name': 'Dim'
            }
        ]

        schema = [
            {
                '_version': 'domain.com/parents-0.1',
                '*': 'https://www.wikidata.org/wiki/Q7565',
                'address': {
                    '*': 'https://www.wikidata.org/wiki/Q319608',
                    'number': {'*': 'https://www.wikidata.org/wiki/Q1413235|lambda _: int(_)'},
                    'street': {'*': 'https://www.wikidata.org/wiki/Q24574749'}
                },
                'children': [
                    {
                        '*': 'https://www.wikidata.org/wiki/Q7569',
                        'age': {'*': 'https://www.wikidata.org/wiki/Q185836|lambda _: float(_)'},
                        'name': {'*': 'https://www.wikidata.org/wiki/Q82799'}
                    }
                ],
                'name': {'*': 'https://www.wikidata.org/wiki/Q82799'}
            }
        ]

        result = [
            {
                'https-www-wikidata-org-wiki-q319608': {
                    'https-www-wikidata-org-wiki-q1413235': 14,
                    'https-www-wikidata-org-wiki-q24574749': 'Leonardo str.'
                },
                'https-www-wikidata-org-wiki-q7569': [
                    {'https-www-wikidata-org-wiki-q185836': 1.0,
                        'https-www-wikidata-org-wiki-q82799': 'Mike'},
                    {'https-www-wikidata-org-wiki-q185836': 15.0,
                        'https-www-wikidata-org-wiki-q82799': 'Tom'}
                ],
                'https-www-wikidata-org-wiki-q82799': 'Max'
            },
            {
                'https-www-wikidata-org-wiki-q319608': {
                    'https-www-wikidata-org-wiki-q1413235': 1,
                    'https-www-wikidata-org-wiki-q24574749': 'Nexus str.'
                },
                'https-www-wikidata-org-wiki-q7569': [
                    {'https-www-wikidata-org-wiki-q185836': 1.0,
                        'https-www-wikidata-org-wiki-q82799': 'Deli'},
                    {'https-www-wikidata-org-wiki-q185836': 7.0,
                        'https-www-wikidata-org-wiki-q82799': 'Miki'}
                ],
                'https-www-wikidata-org-wiki-q82799': 'Dim'
            }
        ]

        self.assertEqual(normalize(data, schema, slugify=True), result)

    def test_make_and_apply_template(self):

        data = {
            'a': [
                {'b': 'c'},
                {'e': 'f'},
                {'g': 'h'}
            ],
            'b': 'something'
        }

        tpl = metaplate(data, ret=True)

        answer = {
            '*': '',
            'a': [
                {'*': '',
                 'b': {'*': ''},
                 'e': {'*': ''},
                 'g': {'*': ''}}
            ],
            'b': {'*': ''}
        }

        self.assertEqual(tpl, answer)

        # apply:

        tpl['a'][0]['b']['*'] = 'hello'

        result = normalize(data, tpl)

        answer = deepcopy(data)

        del answer['a'][0]['b']
        answer['a'][0]['hello'] = 'c'

        self.assertEqual(result, answer)

    def test_make_and_apply_complex_template(self):
        data = {
            'fields': {
                'blockchain': 0,
                'body': '.:en\nThe estimated cost of the laser array is based on extrapolation from the past two decades...',  # noqa
                'categories': [],
                'comment_count': 37,
                'created_date': '2012-10-14T11:43:11',
                'data': None,
                'editors': [],
                'is_draft': False,
                'languages': '["en"]',
                'owner': 120,
                'parents': [32],
                'source': '',
                'title': '.:en:Cost',
                'type': 5,
                'unsubscribed': [],
                'updated_date': '2017-03-21T11:31:53.338'
            },
            'model': 'core.topic',
            'pk': 1
        }

        tpl = metaplate(data, ret=True)

        tpl['fields']['blockchain']['*'] = 'HELLO'

        ndata = normalize(data, tpl, slugify=True)

        answer = deepcopy(data)
        del answer['fields']['blockchain']
        answer['fields']['hello'] = 0

        self.assertEqual(ndata, answer)

    def test_formatization(self):
        ndata = {
            '_:username#string': 'L2174',
            '_:creation-date#unixtime': 1114819200,
            '_:autobiography#string': {
                '_:body-text#string': '\n\n',
                '_:creation-date#isodate': '2005-04-30T00:00:00',
                '_:last-updated#isodate': '2005-05-01T00:00:00'
            },
            '_:idea#object': [
                {'_:url#url': 'http://www.na.com/airbag_20active#1118115294',
                    '_:title#string': 'airbag active'}
            ]
        }

        expect = {
            '_:username': str('L2174'),
            '_:creation-date': datetime.datetime(2005, 4, 30, 0, 0),
            '_:autobiography': {
                '_:body-text': str('\n\n'),
                '_:creation-date': datetime.datetime(2005, 4, 30, 0, 0),
                '_:last-updated': datetime.datetime(2005, 5, 1, 0, 0)},
            '_:idea': [
                {
                    '_:url': 'http://www.na.com/airbag_20active#1118115294',
                    '_:title': str('airbag active')
                }
            ]
        }

        self.assertEqual(formatize(ndata), expect)

    def test_dict_addition_1(self):
        '''
        testing Dict.__add__
        '''
        # It should work for any Dict/List like JSONs.

        a = Dict({'x': 1})
        b = Dict({'x': 1})
        c = Dict({'y': 1})
        d = Dict({'y': [1]})
        e = Dict({'z': {'?': 1}})
        f = Dict({'z': {'?': [1]}})

        assert a + b == {'x': 2}
        assert a + c == {'x': 1, 'y': 1}
        assert a + b + c == {'x': 2, 'y': 1}
        assert d + d == {'y': [1, 1]}
        assert a + d == {'x': 1, 'y': [1]}

        A = Dict({'x': {'y': {'z': 1}, 'u': 2}, 'm': 1, 'n': 2})
        B = Dict({'x': {'y': {'z': 2}, 'u': 8}, 'm': [2], 'n': 3})
        C = Dict({'x': {'y': [{'1': 'Thing A'}, {'2': 'Thing B'}]}})
        D = Dict({'x': {'y': [{'3': 'Thing C'}, {'4': 'Thing D'}]}})

        # Problematic:
        self.assertEqual(c + d, {'y': [1, [1]]})
        # SHOULD BE: {'y': [1,1]}
        self.assertEqual(d + c, {'y': [[1], 1]})
        # SHOULD BE: {'y': [1,1]}
        self.assertEqual(e + f, {'z': [{'?': [1, [1]]}, {'?': [1]}]})
        # SHOULD BE: {'z': {'?': [1,1]]}
        self.assertEqual(A + B,
                         {'x': [{'y': [{'z': 3},
                                       {'z': 2}],
                                 'u': 10},
                                {'y': {'z': 2},
                                 'u': 8}],
                             'm': [1,
                                   [2]],
                             'n': 5})
        # SHOULD BE: {'x': {'y': {'z': 3}, 'u': 10}, 'm': [1,2], 'n': 5}
        self.assertEqual(
            C + D,
            {'x': [{'y': [{'1': 'Thing A', '3': 'Thing C'},
                          {'2': 'Thing B', '4': 'Thing D'},
                          {'3': 'Thing C'},
                          {'4': 'Thing D'}]},
                   {'y': [{'3': 'Thing C'}, {'4': 'Thing D'}]}]}
        )
        # SHOULD BE: {'x': {'y': [{'1': 'Thing A'}, {'2': 'Thing B'}, {'3':
        # 'Thing C'}, {'4': 'Thing D'}]}}

    def test_dict_subtraction_1(self):
        '''
        testing Dict.__sub__
        '''
        a = Dict({'x': 1})
        b = Dict({'x': 1})
        d = Dict({'y': [1]})
        e = Dict({'z': {'?': 1}})
        f = Dict({'z': {'?': [1]}})
        g = Dict({'z': {'?': 1, '!': 1}, 'u': 1})

        empty = {}

        self.assertEqual(b - a, empty)
        self.assertEqual(a - b, empty)
        self.assertEqual(Dict({}) - f, empty)
        self.assertEqual(d - d, {'y': []})

        self.assertEqual(e - e, {'z': {}})

        self.assertEqual(g - e, {'z': {'!': 1}, 'u': 1})

        self.assertEqual(e - f, {'z': {'?': 1}})
        # SHOULD BE: {'z': {'?': []}}


class TestCore(unittest.TestCase):

    def setUp(self):
        self.data = {
            'hello': 1.0,
            'world': 2,
            'how': ['is', {'are': {'you': 'doing'}}]
        }
        with open('metaform/tests/data/topics.json', 'r') as f:
            self.topics = json.load(f)
        with open('metaform/tests/data/comments.json', 'r') as f:
            self.comments = json.load(f)

    def test_generate_template(self):
        expect = {
            '*': '',
            'hello': {'*': ''},
            'how': [{'*': '', 'are': {'you': {'*': ''}}}],
            'world': {'*': ''}
        }
        self.assertEqual(template(self.data), expect)

    def test_rename_keys(self):
        schema = {
            '*': 'greeting',
            'hello': {'*': 'length'},
            'world': {'*': 'atoms'},
            'how': [
                 {
                     '*': 'method',
                     'are': {
                         '*': 'yup',
                         'you': {'*': 'me'}}
                 }
            ]}

        expect = {'atoms': 2,
                  'length': 1.0,
                  'method': ['is', {'yup': {'me': 'doing'}}]}

        self.assertEqual(normalize(self.data, schema), expect)

    def test_apply_lambdas(self):
        schema = {
            '*': 'greeting',
            'hello': {'*': 'length|lambda x: x+5.'},
            'world': {'*': 'atoms|lambda x: str(x)+"ABC"'},
            'how': [
                 {
                     '*': 'method',
                     'are': {
                         '*': 'yup',
                         'you': {'*': 'me|lambda x: "-".join(list(x))'}
                     }
                 }
            ]}

        expect = {
            'atoms': '2ABC',
            'length': 6.0,
            'method': ['is', {'yup': {'me': 'd-o-i-n-g'}}]}

        self.assertEqual(normalize(self.data, schema), expect)

    def test_custom_converters(self):

        def some_func(x):
            a = 123
            b = 345
            return (b - a) * x

        converters.func = some_func

        schema = {
            '*': 'greeting',
            'hello': {'*': 'length|converters.func'},
            'world': {'*': 'atoms|lambda x: str(x)+"ABC"'},
            'how': [
                {
                 '*': 'method',
                 'are': {
                     '*': 'yup',
                     'you': {'*': 'me|lambda x: "-".join(list(x))'}}
                }
            ]}

        expect = {
            'atoms': '2ABC',
            'length': 222.0,
            'method': ['is', {'yup': {'me': 'd-o-i-n-g'}}]}

        self.assertEqual(normalize(self.data, schema), expect)

    def test_merging_topics_and_comments(self):
        topics_schema = [{
            'id': {'*': 'topic-id'},
            'type': {'*': '|lambda x: {0: "NEED", 1: "GOAL", 2: "IDEA", 3: "PLAN", 4: "STEP", 5: "TASK"}.get(x)'},
            'owner': {'username': {'*': ''}, 'id': {'*': 'user-id'}},
            'blockchain': {'*': '|lambda x: x and True or False'},
        }]

        comments_schema = [{
            'id': {'*': 'comment-id'},
            'topic': {'*': 'topic-url'},
            'text': {'*': 'body'},
            'owner': {'username': {'*': ''}, 'id': {'*': 'user-id'}},
            'blockchain': {'*': '|lambda x: x and True or False'},
        }]

        normal_topics = normalize(self.topics, topics_schema)
        normal_comments = normalize(self.comments, comments_schema)

        with open('metaform/tests/data/topics+comments.json', 'r') as f:
            expect = json.load(f)

        self.assertEqual(normal_topics + normal_comments, expect)

    def test_keys_pickout_alignment(self):
        topics_schema = [{
            'id': {'*': 'topic-id'},
            'type': {'*': '|lambda x: {0: "NEED", 1: "GOAL", 2: "IDEA", 3: "PLAN", 4: "STEP", 5: "TASK"}.get(x)'},
            'owner': {'username': {'*': ''}, 'id': {'*': 'user-id'}},
            'blockchain': {'*': '|lambda x: x and True or False'},
        }]
        comments_schema = [{
            'id': {'*': 'comment-id'},
            'topic': {'*': 'topic-url'},
            'text': {'*': 'body'},
            'owner': {'username': {'*': ''}, 'id': {'*': 'user-id'}},
            'blockchain': {'*': '|lambda x: x and True or False'},
        }]

        normal_topics = normalize(self.topics, topics_schema)
        normal_comments = normalize(self.comments, comments_schema)

        abnormal_comments = [
            dict(comment, **{"some": {"place": {"deep": comment["owner"]}}, "owner": None})
            for comment in normal_comments
        ]

        with open('metaform/tests/data/topics+comments-pickout.json', 'r') as f:
            expect = json.load(f)

        self.assertEqual(
            json.loads(json.dumps(list(
                align([normal_topics[:1], abnormal_comments[:1]])
            ))), expect
        )

    def test_keys_renaming(self):
        self.assertEqual(
            normalize({'A': 1}, {'A': {'*': 'B'}}),
            {'B': 1}
        )

    def test_hyphenation(self):
        self.assertEqual(
            normalize({'A': 1}, {'A': {'*': 'a_b'}}),
            {'a_b': 1}
        )

    def test_keys_renaming_slugify(self):
        self.assertEqual(
            normalize({'A': 1}, {'A': {'*': 'B'}}, slugify=True),
            {'b': 1}
        )

    def test_hyphenation_slugify(self):
        self.assertEqual(
            normalize({'A': 1}, {'A': {'*': 'a_b'}}, slugify=True),
            {'a-b': 1}
        )


if __name__ == '__main__':
    unittest.main()



import unittest


class TestUtils(unittest.TestCase):

    def setUp(self):
        self.test = 'Yes'

    def test_simple_conversion(self):
        result = 1
        expect = 1

        self.assertEqual(result, expect)


if __name__ == '__main__':
    unittest.main()



import operator
import os
import pathlib
from collections import defaultdict
from copy import deepcopy
from functools import reduce

import metawiki
import yaml
from boltons.iterutils import remap
from tinydb import Query, TinyDB
from typology import Concept
from typology.utils import get_schema as t_get_schema  # noqa
from typology.utils import slug

conf_path = os.path.join(str(pathlib.Path.home()), '.metaform')

if not os.path.exists(conf_path):
    os.makedirs(conf_path)

concepts = TinyDB(os.path.join(conf_path, 'concepts.json'))
schemas = TinyDB(os.path.join(conf_path, 'schemas.json'))


def dictget(d, mapList):
    '''
    Given:
    >>> d = {'a': [{'b': 'c'}, {'e': 'f'}, {'g': 'h'}]}
    Returns:
    >>> dictget(d, ['a', 2, 'b'])
    >>> 'c'
    '''
    return reduce(operator.getitem, mapList, d)


def metapath(path):
    '''
    Given:
    >>> p = [1, 'a', 2, 'b', 3]
    Returns:
    >>> metapath(p)
    >>> [0, 'a', 0, 'b', 0]
    '''
    metapath = []
    for item in path:
        if isinstance(item, int):
            metapath.append(0)
        else:
            metapath.append(item)
    return metapath


def metaplate(data, with_self=True, print_yaml=False):
    '''
    Creates a meta-template for a dictionary-like data.

    Given:
    >>> data = {'a': [{'b': 'c'}, {'e': 'f'}, {'g': 'h'}], 'b': 'something'}

    Returns template:
    >>> metaplate(data) == result
    result = {
        '*': '',
        'a': [
            {'*': '',
             'b': {'*': ''},
             'e': {'*': ''},
             'g': {'*': ''}}
        ],
        'b': {'*': ''}
    }
    '''
    if isinstance(data, list):
        # TODO: reserved key
        data = {'_#list#_': data}

    def sum_dicts(list_of_dicts):
        result = {}
        for item in list_of_dicts:
            if isinstance(item, dict):
                result.update(item)
        return [result]

    def visit(path, key, value):
        if isinstance(value, list):
            combined = sum_dicts(value)
            combined[0].update({'*': ''})
            return key, combined
        if not isinstance(value, list) and not isinstance(value, dict):
            return key, {'*': ''}
        else:
            return key, value

    remapped = remap(data, visit=visit)

    if with_self:
        if isinstance(remapped, dict):
            remapped.update({'*': ''})
        elif isinstance(remapped, list):
            if remapped:
                if remapped[0]:
                    remapped[0].update({'*': ''})

    if '_#list#_' in remapped.keys():
        remapped = remapped['_#list#_']

    if print_yaml:
        print(yaml.dump(remapped))
    else:
        return remapped


def get_schema(path, refresh=False):

    url = metawiki.name_to_url(str(path))

    if any(
        [str(url).startswith(it) for it in
         list(metawiki.NAMESPACES.keys()) + ['https://github.com', 'https://www.wikidata.org']]
    ):

        Schemas = Query()

        slg = slug(url)

        result = schemas.search(Schemas.slug == slg)

        if not result or refresh:

            if refresh:

                try:
                    schemas.remove(Schemas.slug == slg)
                except BaseException:
                    pass

            try:
                schema = t_get_schema(url)
                result = {'slug': slg, 'schema': schema}
                schemas.insert(result)

                return schema

            except BaseException:
                print("-> Could not find schema: {}".format(url))

        else:
            return result[0]['schema']


def get_concept(value, refresh=False):

    url = metawiki.name_to_url(str(value))

    if any(
        [str(url).startswith(it) for it in
         list(metawiki.NAMESPACES.keys()) + ['https://github.com', 'https://www.wikidata.org']]
    ):

        Concepts = Query()

        slg = slug(url)

        result = concepts.search(Concepts.slug == slg)

        if not result or refresh:

            try:
                concept = Concept(url).concept
                result = {'slug': slg, 'concept': concept}
                concepts.insert(result)

                return concept

            except BaseException:
                print("-> Undefined concept: {}".format(url))

        elif result:
            return result[0]['concept']
    else:
        return


def get_concept_paths(data, k=[], exclude=[dict, list]):
    '''
    Given, something like:
    >>> get_concept_paths({'a': {'b': [{'c': 1}]}, 'x': {'c': 2}}, 'c')

    Returns a list of paths, where given concept ('c') is
    >>> [['a', 'b', 0, 'c'], ['x', 'c']] # = p

    So that we can retrieve them with dictget()
    '''

    if isinstance(k, list):
        paths = {key: [] for key in k}
    else:
        paths = []

    if not k:
        paths = defaultdict(list)

    def visit(path, key, value):
        if isinstance(k, list):
            if key in k or not k:
                if type(value) not in exclude:
                    paths[key].append(list(path) + [key])
        else:
            if key == k or not k:
                if type(value) not in exclude:
                    paths.append(list(path) + [key])
        return key, value

    remap(data, visit=visit)

    return paths


def get_match_matrix(
        source_schema_list, exclude=[dict, list]):
    '''
    Given a few source schemas, e.g.
    metaform.utils.match_sources([ {'a': {'c': 'X'}, 'n': 1}, {'b': {'a': {'c': 'Y'}}, 'd': {'n': 2}} ])
    >>> get_matche_matrix([{'a': {'b': 'X'}, 'n': 1}, {'b': {'a': {'c': 'Y'}}, 'd': {'n': 2}}])

    >>> {'n': [['n'], ['d', 'n']]}
    >>> metaform.utils.get_match_matrix([ {'a': {'c': 'X'}, 'n': 1}, {'b': {'a': {'c': 'Y'}}, 'd': {'n': 2}} ])
    >>>

    Returns found matching keys, that can be merged.
    '''

    field_availability_list = [
        get_concept_paths(source, exclude=exclude)
        for source in source_schema_list
    ]

    common_field_set = set.intersection(*map(set, field_availability_list))

    match = {}
    for field in common_field_set:
        lists = []
        for source in field_availability_list:
            lists.append(source[field][0])
        match[field] = lists

    return match


def match(dict_list, exclude=[dict, list]):

    results = {}

    matches = get_match_matrix(dict_list, exclude=exclude)

    for key in matches:
        results.update(
            {key: [dictget(source, matches[key][i])
                   for i, source in enumerate(dict_list)]})

    return results


def getx(data, path, inany=False):
    if not inany:
        try:
            return dictget(data, path)
        except BaseException:
            return None
    else:
        oneup = dictget(data, path[:-1])
        if isinstance(oneup, list):
            possible_paths = [path[:-1] + (i,) + path[-1:] for i in range(len(oneup))]
            for p in possible_paths:
                try:
                    item = dictget(data, p)
                    break
                except BaseException:
                    item = None
            if item is not None:
                return p, item
            else:
                return None


def setx(data, path, value, other):
    '''
    Example:
    >>> setx({'a': 1, 'b': 2}, ['c', 0, 'd'], 8, {'c': [{'d': 12}]})
    {'a': 1, 'b': 2, 'c': [{'d': 8}]}
    '''
    r = data
    i = []
    for p in path[:-1]:
        i += [p]

        if (isinstance(r, dict) and (p not in r.keys())) or \
           (isinstance(r, list) and (p not in range(len(r)))):
            try:
                other_type = type(getx(other, i))  # [], {}
            except BaseException:
                other_type = dict

            if isinstance(r, dict):
                r[p] = other_type()
            elif isinstance(r, list):
                r += [other_type()]
            else:
                print('hmm')

        r = r[p]

    try:
        r[path[-1]] = value
    except BaseException:
        # do nothing #
        pass

    return data


def _add(a, b):
    a = deepcopy(a)

    def visit(path, key, value):
        fpath = path + (key,)
        val = getx(a, fpath)

        if val is not None:
            if not isinstance(val, type(value)):
                new_val = [val, value]
            else:
                # If upper element is concatable, don't add item values per se.
                if type(getx(a, fpath[:-1])) in [list, tuple]:
                    if type(getx(b, fpath[:-1])) in [list, tuple]:
                        return key, value

                if hasattr(val, '__add__'):
                    new_val = val + value
                else:
                    if val == value:
                        new_val = val
                    else:
                        new_val = [val, value]

            setx(a, fpath, new_val, b)
        else:
            setx(a, fpath, value, b)

        return key, value

    remap(b, visit=visit)
    return a


def delx(data, path):
    '''
    >>> delx({'a': 3, 'b': [2, {'x': 'y'}], 'c': 3, 'd': 4}, ['b', 1, 'x'])
    {'a': 3, 'b': [2, {}], 'c': 3, 'd': 4}
    >>> delx({'a': 3, 'b': [2, {'x': 'y'}], 'c': 3, 'd': 4}, ['b', 0])
    '''
    # data = deepcopy(data)

    r = data
    for p in path[:-1]:
        r = r[p]
    del r[path[-1]]
    if not r:
        del r

    return data


def _sub(a, b):
    '''
    >>> _sub({'a': 3, 'b': [2, {'x': 'y'}], 'c': 3, 'd': 4},
             {'a': 2, 'b': {'x': 'y'}, 'c': 3, 'd': 4})

    {'a': 1, 'b': 2}
    '''
    a = deepcopy(a)
    a['___previous_was_list___'] = False

    def visit(path, key, value):
        fpath = path + (key,)
        vala = getx(a, fpath)
        valb = value

        if isinstance(getx(a, fpath[:-1]), list):
            valA = getx(a, fpath, inany=True)
        else:
            valA = None

        if vala is not None:
            if not isinstance(vala, type(valb)):
                if not isinstance(vala, list):
                    vala = [vala]
                if not isinstance(valb, list):
                    valb = [valb]
                new_val = [v for v in vala if v in valb]
                if len(new_val) == 1:
                    new_val = new_val[0]
                if not a['___previous_was_list___']:
                    setx(a, fpath, new_val, b)
                a['___previous_was_list___'] = False
                # print('>>', a, (vala, valb))
            elif hasattr(vala, '__sub__'):
                new_val = vala - valb
                if not a['___previous_was_list___']:
                    if new_val:
                        setx(a, fpath, new_val, b)
                    else:
                        delx(a, fpath)

        else:
            if valA is not None:
                # print(valA)
                npath, _ = valA
                # print('->', a)
                delx(a, npath[:-1])
                # print('-<', a)
                a['___previous_was_list___'] = True
        return key, value

    remap(b, visit=visit)
    del a['___previous_was_list___']
    return a



# Copyright (c) 2018 WeFindX Foundation, CLG.
# All Rights Reserved.

from setuptools import find_packages, setup

with open('README.rst', 'r') as f:
    long_description = f.read()


setup(
    name='metaform',
    version='1.0.1.1',
    description='A utility for defining metadata for data types and formats.',
    long_description=long_description,
    url='https://gitlab.com/wefindx/metaform',
    author='Mindey',
    author_email='mindey@qq.com',
    license='ASK FOR PERMISSIONS',
    packages=find_packages(exclude=['docs', 'tests*']),
    install_requires=[
        "metawiki",
        "typology",
        "metadir",
        "tinydb",
        "python-dateutil",
    ],
    extras_require={
        'develop': [
            'pre-commit==1.18.3',
            'coverage==4.5.4',
            'flake8==3.7.8',
            'isort==4.3.21',
        ],
    },
    zip_safe=False
)
